/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef _THREADS_H_
#define _THREADS_H_

#include "lists.h"

#ifndef NULL
# define NULL 0
#endif

#define PRIO_MIN (-20)
#define PRIO_MAX   20

/* Processor's own stack pointer. It's not an actual
 * variable and can be changed only with assembly. */
extern void *sp;

extern unsigned int clock;

typedef struct thread_t {
	node_t node;
	void *(*code)(void);
	void *sp;
	void *stack_top;
	enum thread_state {STATE_RUNNING, STATE_SUSPENDED, STATE_STARTED, STATE_EXITED, STATE_CANCELLED} state;
	void *retval;
	signed char nice;
	unsigned char ticks_int; /* integer part of ticks per timeslice */
	unsigned char ticks_frac; /* fraction part of ticks per timeslice */
	unsigned char ticks_accum; /* accumulator for ticks_frac */
} thread_t;

typedef struct userint_t {
	void (*code)(void);
	struct userint_t *next;
} userint_t;

typedef struct userintlist_t {
	struct userint_t *head;
} userintlist_t;

void threads_init(void);
int thread_create(thread_t *t, void *(*code)(void), void *stack_top);
int thread_nice(thread_t *t, int inc);
int thread_cancel(thread_t *t);
void thread_suspend(void);
void thread_yield(void);
void thread_exit(void *retval);
void *thread_join(thread_t *t);

#endif /* _THREADS_H_ */
