; threadsp, functions for preemptive multithreading
; Copyright 2002, 2003, 2004 Chris Williams <cow3.14@juno.com>
; 
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; To include optional function foo, for example, define NEED_thread_foo.
; Optional functions: cancel, suspend, join.

PRIO_MIN	= -20
PRIO_MAX	= 20

STATE_RUNNING	= 0
STATE_SUSPENDED	= 1
STATE_STARTED	= 2
STATE_EXITED	= 3
STATE_CANCELLED	= 4


 ifdef NEED_thread_join
thread_t.retval.sizeof = 2
 else
thread_t.retval.sizeof = 0
 endif
 ifdef NEED_thread_nice
nicestuff.sizeof = 4
 else
nicestuff.sizeof = 0
 endif

thread_t.sizeof		= 9+thread_t.retval.sizeof+nicestuff.sizeof
thread_t.node		= 0
thread_t.node.next	= 0
thread_t.code		= 2
thread_t.sp		= 4
thread_t.stack_top	= 6
thread_t.state		= 8
thread_t.nice		= 9
thread_t.ticks_int	= 10
thread_t.ticks_frac	= 11
thread_t.ticks_accum	= 12
thread_t.retval		= 9+nicestuff.sizeof

list_t.sizeof	= 4
list_t.head	= 0
list_t.tail	= 2

node_t.sizeof	= 2
node_t.next	= 0

 define TIMESLICE_LENGTH 4

; define SIM_ONLY ; don't change the real stack pointer (SP)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; push_all and pop_all
; pushes or pops all registers except IY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 define push_all push af\ push bc\ push de\ push hl\ push ix
 define pop_all pop ix\ pop hl\ pop de\ pop bc\ pop af

 define cpo(n) #if n\ cp n\ #else\ or a\ #endif
 define ldao(n) #if n\ ld a,n\ #else\ xor a\ #endif
;#define nextmembbc(old,new) nextmemb(bc,old,new)
;#define nextmembde(old,new) nextmemb(de,old,new)
;#define nextmemb(r,old,new) #if (new)-(old) > 3 || (old)-(new) > 3\ ld r,(new)-(old)\ add hl,r\ #else\ addhl((new)-(old))\ #endif
;#define addhl(n) #if n > 0\ inc hl\ #endif\ #if n > 1\ inc hl\ #endif\ #if n > 2\ inc hl\ #endif\ #if -n > 0\ dec hl\ #endif\ #if -n > 1\ dec hl\ #endif\ #if -n > 2\ dec hl\ #endif

 module thread
	ld	hl,0
	add	hl,sp
	ld	(_exitsp),hl	; save the stack pointer
	call	_xxx			; so exit resets the int mode
	im	1
exit:
	ld	hl,(_exitsp)
	ld	sp,hl		; restore the stack pointer
	ret
_xxx
	ld	hl,0
	ld	(_threads+list_t.head),hl
	ld	(_threads+list_t.tail),hl	; threads.head = threads.tail = NULL
	ld	(clock),hl		; clock = 0;
	ld	d,h
	ld	e,l
	add	hl,sp
	ld	b,h
	ld	c,l
	ld	hl,_main_thread
	call	thread_create		; thread_create(&main_thread, (void *)NULL, sp + sizeof(void *))
	ldao(STATE_RUNNING)
	ld	(_main_thread+thread_t.state),a	; main_thread.state = STATE_RUNNING
	ld	a,1
	ld	(_num_running_threads),a	; num_running_threads = 1
	ld	a,TIMESLICE_LENGTH
	ld	(_timeslice_ticks),a	; timeslice_ticks = TIMESLICE_LENGTH
	call	int_init		; int_init()
	jp	_end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread to create
; 	DE	address of code
; 	BC	top address of stack
; output:
; 	Carry set on error
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_create:
	ld	a,l
	or	h			; !t
	scf
	ret	z			; return -1
		push	hl
 if thread_t.code
 if thread_t.code == 1
	inc	hl
 else
		push	de
	ld	de,thread_t.code
	add	hl,de
		pop	de
 endif
 endif
	ld	(hl),e
	inc	hl
	ld	(hl),d			; t->code = code
 if thread_t.stack_top-thread_t.code-1 == 1
	inc	hl
 else
	ld	de,thread_t.stack_top-thread_t.code-1
	add	hl,de
 endif
	ld	(hl),c
	inc	hl
	ld	(hl),b			; t->stack_top = stack
 if thread_t.state-thread_t.stack_top-1 == 1
	inc	hl
 else
	ld	de,thread_t.state-thread_t.stack_top-1
	add	hl,de
 endif
	ld	(hl),STATE_STARTED	; t->state = STATE_STARTED
 ifdef NEED_thread_nice
 if thread_t.nice-thread_t.state == 1
	inc	hl
 else
	ld	de,thread_t.nice-thread_t.state
	add	hl,de
 endif
	ld	(hl),0			; t->nice = 0
 if thread_t.ticks_accum-thread_t.nice == 1
	inc	hl
 else
	ld	de,thread_t.ticks_accum-thread_t.nice
	add	hl,de
 endif
	ld	(hl),256/2		; t->ticks_accum = 256/2
		pop	hl
		push	hl
	xor	a
	call	thread_nice		; thread_nice(t, 0)
 endif ; NEED_thread_nice
		pop	hl
 if thread_t.node
	ld	de,thread_t.node
	add	hl,de
 endif
	ex	de,hl
	ld	hl,_threads
		di
	call	list_append_node	; list_append_node(&threads, &t->node)
		ei
	;call	thread_yield		; thread_yield()
	or	a
	ret				; return 0

 ifdef NEED_thread_nice
; FIXME: this is pretty bad
; I'll wait 'til this function in threads.c cleans up and
; stabilizes a bit more before I clean this up here.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread
; 	A	increment for nice
; output:
; 	Carry set on error?
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_nice:
	ld	e,a
	ld	a,l
	or	h
	jr	nz,_nice0		; !t
	ld	hl,(_threads+list_t.head)	; t = (thread_t *)threads.head
_nice0
	ld	a,e
	rla
	sbc	a,a
	ld	d,a	; sext A to DE
	ld	bc,thread_t.nice
	add	hl,bc
		di			; intoff()
	ld	a,(hl)
		ei
	ld	c,a
	rla
	sbc	a,a
	ld	b,a	; sext A to BC
	ex	de,hl
	add	hl,bc
	ld	bc,-PRIO_MIN
	add	hl,bc			; nice = t->nice + inc - PRIO_MIN
	ld	a,l
	bit	7,h
	jr	z,_nice1		; nice < 0
	ldao(PRIO_MIN)			; nice = PRIO_MIN
	jr	_nice3
_nice1					; else
	cp	PRIO_MAX-PRIO_MIN+1
	jr	c,_nice2		; nice > (PRIO_MAX - PRIO_MIN)
	ldao(PRIO_MAX)			; nice = PRIO_MAX
	jr	_nice3
_nice2
	add	a,PRIO_MIN
_nice3
	ex	de,hl
		di
	ld	(hl),a			; t->nice = nice
	neg	; -nice
	ld	d,a
	xor	a
	sra	d\ rra	; divide -nice by 2...
	sra	d\ rra	; ...4...
	sra	d\ rra	; ...8
	ld	e,a
 if thread_t.ticks_int-thread_t.nice == 1
	inc	hl
 else
	ld	bc,thread_t.ticks_int-thread_t.nice
	add	hl,bc	; hl = &t->ticks_int
 endif
	ldao(TIMESLICE_LENGTH)
	add	a,d
	ld	(hl),a			; t->ticks_int = TIMESLICE_LENGTH + -nice / 8
 if thread_t.ticks_frac-thread_t.ticks_int == 1
	inc	hl
 else
	ld	bc,thread_t.ticks_frac-thread_t.ticks_int
	add	hl,bc	; hl = &t->ticks_frac
 endif
	ld	(hl),e			; t->ticks_frac = -nice * 32
		ei			; inton()
	or	a
	ret				; return 0
 endif ; NEED_thread_nice

 ifdef NEED_thread_cancel
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread to cancel
; output:
; 	Carry set on error
; registers destroyed:
; 	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_cancel:
	ld	a,l
	or	h			; !t
	jr	nz,_cancel0
	ld	hl,(_threads+list_t.head)	; t = (thread_t *)threads.head
_cancel0
	ld	de,-_main_thread
	ex	de,hl
	add	hl,de			; t == &main_thread
	ld	a,l
	or	h
	jp	z,exit			; exit(0);
_cancel1
 if thread_t.state
	ld	hl,thread_t.state
	add	hl,de
 else
	ex	de,hl
 endif
		di
	ld	a,(hl)
	cpo(STATE_RUNNING)
	jr	nz,_cancel2
	ex	de,hl
	ld	hl,_num_running_threads
	dec	(hl)
	ex	de,hl			; --num_running_threads
_cancel2
	ld	(hl),STATE_CANCELLED	; t->state = STATE_CANCELLED
	or	a
	jp	thread_yield		; thread_yield(); return 0
 endif

 ifdef NEED_thread_join
 define NEED_thread_suspend
 endif
 ifdef NEED_thread_suspend
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	none
; output:
; 	none
; registers destroyed:
; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_suspend:
	ld	hl,(_threads+list_t.head)
 if thread_t.state
	ld	de,thread_t.state
	add	hl,de			; s = &((thread_t *)threads.head)->state
 endif
		di
	ld	(hl),STATE_SUSPENDED	; *s = STATE_SUSPENDED
	ld	hl,_num_running_threads
	dec	(hl)			; --num_running_threads
	jp	thread_yield		; thread_yield()
 endif	; NEED_thread_suspend

 ifdef NEED_thread_nice
; input:
; 	hl = thread
; output:
; 	a = length of timeslice, in ticks
; 	zero flag is set if result is zero
; 	_timeslice_ticks is set too
_settimeslice:
 if thread_t.ticks_int
	ld	bc,thread_t.ticks_int
	add	hl,bc
 endif
	ld	a,(hl)
	ld	de,_timeslice_ticks
	ld	(de),a		; timeslice_ticks = t->ticks_int
 if thread_t.ticks_frac-thread_t.ticks_int == 1
	inc	hl
 else
	ld	bc,thread_t.ticks_frac-thread_t.ticks_int
	add	hl,bc
 endif
	ld	a,(hl)
 if thread_t.ticks_accum-thread_t.ticks_frac == 1
	inc	hl
 else
	ld	bc,thread_t.ticks_accum-thread_t.ticks_frac
	add	hl,bc
 endif
	add	a,(hl)
	ld	(hl),a			; t->ticks_accum += t->ticks_frac
	ld	a,(de)
	adc	a,0			; t->ticks_accum < t->ticks_frac
	ld	(de),a			; ++timeslice_ticks
	ret
 endif

 include "ints.inc"
int_start:
;	jr	$+4
;	reti
	exx
	ex	af,af'
	call	$0066	; default TI-86 int handler
	
	exx
	ex	af,af'
	ld	hl,(clock)
	inc	hl
	ld	(clock),hl		; ++clock
	ld	hl,_timeslice_ticks
	dec	(hl)			; --timeslice_ticks
	exx
	jr	z,_int_start1
	ex	af,af'
	ret
_int_start1
 ifndef NEED_thread_nice
	ldao(TIMESLICE_LENGTH)
	ld	(_timeslice_ticks),a	; timeslice_ticks = TIMESLICE_LENGTH
 endif
	ex	af,af'

 define thread_yield thread_switch
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	none
; output:
; 	none
; registers destroyed:
; 	none
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_switch:
		di
		push_all
 ifdef SIM_ONLY
	ld	a,(_threads+list_t.head+1)
	call	puthex
	ld	a,(_threads+list_t.head)
	call	puthex
	ld	a,' '
	call	putc
	ld	a,(_sp+1)
	call	puthex
	ld	a,(_sp)
	call	puthex
 endif
 ifdef SIM_ONLY
	ld	de,(_sp)
 else
	ld	hl,0
	add	hl,sp
	ex	de,hl
 endif
	ld	hl,(_threads+list_t.head)
 if thread_t.sp
	ld	bc,thread_t.sp
	add	hl,bc
 endif
	ld	(hl),e
	inc	hl
	ld	(hl),d			; ((thread_t *)threads.head)->sp = sp
_switch0	; next:
	ld	hl,(_threads+list_t.head)
	ld	(_threads+list_t.tail),hl	; threads.tail = threads.head
;	ld	hl,(_threads+list_t.head)
 if thread_t.node.next
	ld	bc,thread_t.node.next
	add	hl,bc
 endif
	call	LD_HL_MHL
	ld	(_threads+list_t.head),hl	; threads.head = threads.head->next
_switch1				; while (1) {
	ld	hl,(_threads+list_t.head)
 if thread_t.state
	ld	bc,thread_t.state
	add	hl,bc
 endif
	ld	a,(hl)			; switch (t->state) {
	cpo(STATE_RUNNING)		; case STATE_RUNNING
	jr	nz,_switch2
;_running	; HL must be (_threads+list_t.head)+thread_t.state
 ifdef NEED_thread_nice
		push	hl
	ld	bc,-thread_t.state
	add	hl,bc
	call	_settimeslice		; settimeslice(t)
		pop	hl
	jr	z,_switch0		; goto next
 endif ; NEED_thread_nice
	ld	bc,thread_t.sp-thread_t.state
	add	hl,bc
	call	LD_HL_MHL
 ifdef SIM_ONLY
	ld	(_sp),hl
 else
	ld	sp,hl			; sp = ((thread_t *)threads.head)->sp
 endif
		pop_all
		ei
	ret	; return to a running thread
_switch2
	cpo(STATE_SUSPENDED)		; case STATE_SUSPENDED
	jr	nz,_switch3
	
	ex	de,hl
	ld	bc,(_threads+list_t.head)
	ld	hl,-_main_thread
	add	hl,bc
	ld	a,l
	or	h
	ex	de,hl
	jr	nz,_switch6	; (thread_t *)threads.head = &main_thread
	ld	a,(_num_running_threads)
	or	a			; num_running_threads == 0
	jr	nz,_switch6
	
	ld	(_timeslice_ticks),a	; timeslice_ticks = 0
	
		ei
	halt
		di
_switch6
	ld	(hl),STATE_RUNNING	; t->state = STATE_RUNNING;
	ld	hl,_num_running_threads
	inc	(hl)			; ++num_running_threads;
	jr	_switch1		; break
_switch3
	cpo(STATE_STARTED)		; case STATE_STARTED
	jr	nz,_switch5
 ifdef NEED_thread_nice
	ld	bc,-thread_t.state
	add	hl,bc
	call	_settimeslice		; settimeslice(t)
	jr	z,_switch0		; goto next
 endif ; NEED_thread_nice
	ld	hl,(_threads+list_t.head)
	ld	d,h
	ld	e,l			; de = t
 if thread_t.stack_top
	ld	bc,thread_t.stack_top
	add	hl,bc
 endif
	call	LD_HL_MHL
 ifdef SIM_ONLY
	ld	(_sp),hl
 else
	ld	sp,hl			; sp = t->stack_top
 endif
	ld	hl,_num_running_threads
	inc	(hl)			; ++num_running_threads
 if thread_t.state
	ld	hl,thread_t.state
	add	hl,de
 else
	ld	h,d
	ld	l,e
 endif
	ld	(hl),STATE_RUNNING	; t->state = STATE_RUNNING
	ld	bc,thread_t.code-thread_t.state
	add	hl,bc
	call	LD_HL_MHL
 ifdef SIM_ONLY
	call	thread_switch
 else
		ei
	call	JP_MHL			; t->code()
	jr	thread_exit		; thread_exit(...)
 endif
_switch5	; default
	ld	hl,_threads
	call	list_remove_head	; list_remove_head(&threads)
	jp	_switch1

JP_MHL:	; call this to call a function whose address is in HL
	jp	(hl)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	exit status
; output:
; 	none (does not return)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_exit:
	ex	de,hl
	ld	hl,(_threads+list_t.head)	; t = (thread_t *)threads.head
	ld	bc,-_main_thread
		push	hl
	add	hl,bc
	ld	a,l
	or	h			; t == &main_thread
		pop	hl
	jp	z,exit			; exit(0);
		di			; intoff()
 if thread_t.state
	ld	bc,thread_t.state
	add	hl,bc
 endif
	ld	(hl),STATE_EXITED	; t->state = STATE_EXITED
 ifdef NEED_thread_join
	ld	bc,thread_t.retval-thread_t.state
	add	hl,bc
	ld	(hl),e
	inc	hl
	ld	(hl),d			; t->retval = status
 endif
	ld	hl,_num_running_threads
	dec	(hl)			; --num_running_threads
	jp	thread_yield		; thread_yield()

 ifdef NEED_thread_join
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input:
; 	HL	address of thread to join
; output:
; 	Carry is clear if the thread exitted
; 	Carry is set if the thread was cancelled
; 	HL = thread's return value (only if it exitted)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
thread_join:
	ld	a,l
	or	h			; !t
	scf
	ret	z			; return -1
 if thread_t.state
	ld	de,thread_t.state
	add	hl,de
 endif
_join1					; while (1) {
		di
	ld	a,(hl)
	cpo(STATE_EXITED)		; t->state != STATE_EXITED
	jr	z,_join2
	cpo(STATE_CANCELLED)		; t->state != STATE_CANCELLED
	jr	z,_join3
		push	hl
	call	thread_suspend
		pop	hl
	jr	_join1
_join2
	ld	de,thread_t.retval-thread_t.state
	add	hl,de
	call	LD_HL_MHL
		ei
	or	a
	ret				; return t->retval
_join3
		ei
		pop	hl
	scf
	ret				; return -1
 endif

	dseg	; should this be xseg?
_exitsp:
	ds	2
_threads				; list_t threads
	ds	list_t.sizeof
_main_thread:				; thread_t main_thread
	ds	thread_t.sizeof
_num_running_threads:			; int num_running_threads
	ds	1
_timeslice_ticks:
	ds	1
clock:
	ds	2
 ifdef SIM_ONLY
_sp:	ds	2
 endif
	cseg

 include "lists.inc"
 module noname

 ifdef SIM_ONLY
 include "puthex.inc"
 endif

 module thread
_end
 module noname
